#include <gmsh.h>
#include <iostream>
#include <map>
#include <sstream>
#include <Eigen/Dense>
#include <cmath>
#include <iomanip>
#include <stdio.h>
#include <chrono>
#include "functionsBEM.hpp"
#include <omp.h>
#include <algorithm>
#include <list>

void getBC(std::map<std::string, std::pair<bool, double>> &bcMap,
           const std::string domainName)
{
    // Gets the boundary conditions from onelab and fills a map: "name of the physical group" -> (true/false, value).
    // true: dirichlet boundary condition
    // false: neumann boundary condition
    std::vector<std::string> bcKeys;
    gmsh::onelab::getNames(bcKeys, "Boundary Conditions.+");

    for(auto &key : bcKeys)
    {
        std::vector<double> values;
        gmsh::onelab::getNumber(key, values);

        std::stringstream ss(key);
        std::vector<std::string> words;
        std::string word;
        while(getline(ss, word, '/'))
            words.push_back(word);

        // Only focus on the BEM boundary conditions.
        if(words.size() == 4 && words[2].compare(domainName) == 0)
        {
            if(words[3].compare("dirichlet") == 0)
                bcMap[words[1]] = {true, values[0]};
            else if(words[3].compare("neumann") == 0)
                bcMap[words[1]] = {false, values[0]};
        }
    }
}

std::size_t fillElementVector(std::vector<elementStruct> &elementVector,
                      std::map<std::string, std::pair<int, int>> &physicalGroups,
                      std::map<int, std::pair<double, double>> &nodalDisplacements,
                      const std::string domainName)
{
    std::size_t nbDirichletElements = 0;
    
    std::map<std::string, std::pair<bool, double>> bcMap;
    getBC(bcMap, domainName);

    // Gets the dim and the tags of the physical groups.
    gmsh::vectorpair physicalGroupsDimTags;
    gmsh::model::getPhysicalGroups(physicalGroupsDimTags);

    for(std::size_t i = 0; i < physicalGroupsDimTags.size(); i++)
    {
        // Fills the "physicalGroups" map that associates the name to the {dim, tag}.
        int physicalGroupDim = physicalGroupsDimTags[i].first;
        int physicalGroupTag = physicalGroupsDimTags[i].second;
        std::string physicalGroupName;
        gmsh::model::getPhysicalName(physicalGroupDim, physicalGroupTag, physicalGroupName);
        physicalGroups[physicalGroupName] = {physicalGroupDim, physicalGroupTag};

        gmsh::model::getPhysicalName(physicalGroupDim, physicalGroupTag, physicalGroupName);
        
        // If no boundary condition on a given physical curve, don't do anything.
        if(bcMap.find(physicalGroupName) == bcMap.end())
            continue;

        // Retrieves the type and the value of the boundary condition of this physical group.
        // /!\ Should check that this physical groups was well defined in the list of "Boundary Condition" !
        bool bcType = bcMap[physicalGroupName].first;
        double bcValue = bcMap[physicalGroupName].second;

        std::vector<int> entityTags;
        gmsh::model::getEntitiesForPhysicalGroup(physicalGroupDim, physicalGroupTag, entityTags);
        for (std::size_t i = 0; i < entityTags.size(); i++)
        {
            std::vector<int> elementTypes = {1, 8};
            for(std::size_t j = 0; j < elementTypes.size(); j++)
            {
                int nbNodesPerElement = 2;
                if(elementTypes[j] == 8)
                    nbNodesPerElement = 3;

                std::vector<std::size_t> elementTags;
                std::vector<std::size_t> elementNodeTags;
                gmsh::model::mesh::getElementsByType(elementTypes[j], elementTags, elementNodeTags, entityTags[i]);

                std::vector<std::size_t> nodeTags;
                std::vector<double> nodeCoords;
                std::vector<double> parametricNodeCoords;
                gmsh::model::mesh::getNodesByElementType(elementTypes[j], nodeTags, nodeCoords, parametricNodeCoords, entityTags[i], false);

                std::map<std::size_t, std::size_t> nodeIndices;
                for(std::size_t k = 0; k < nodeTags.size(); k++)
                    nodeIndices[nodeTags[k]] = k;
                
                #pragma omp parallel shared(nbDirichletElements, elementTags, nodeTags, nodalDisplacements, nodeIndices, nodeCoords, bcType, bcValue)
                {
                    std::vector<elementStruct> localDirichletElementVector;
                    std::vector<elementStruct> localNeumannElementVector;

                    #pragma omp for schedule(guided)
                    for(std::size_t el = 0; el < elementTags.size(); el++)
                    {
                        std::size_t elTag = elementTags[el];

                        elementStruct element;
                        element.tag = elTag;
                        element.nodeTags[0] = nodeTags[nbNodesPerElement*el];
                        element.nodeTags[1] = nodeTags[nbNodesPerElement*el + 1];

                        if(nbNodesPerElement == 3)
                            element.midNodeTag = nodeTags[nbNodesPerElement*el + 2];
                        else
                            element.midNodeTag = 0;
                        
                        std::pair<double, double> nodalDisp1 = {0.0, 0.0}, nodalDisp2 = {0.0, 0.0};
                        if(nodalDisplacements.count(element.nodeTags[0]))
                            nodalDisp1 = nodalDisplacements[element.nodeTags[0]];
                        if(nodalDisplacements.count(element.nodeTags[1]))
                            nodalDisp2 = nodalDisplacements[element.nodeTags[1]];
                        computeGeometricParam(element, nodeIndices, nodeCoords, nodalDisp1, nodalDisp2);

                        if(bcType)
                        {
                            element.bcValue[0] = bcValue;
                            localDirichletElementVector.push_back(element);
                        }
                        else
                        {
                            element.bcValue[1] = bcValue;
                            localNeumannElementVector.push_back(element);
                        }
                    }

                    #pragma omp critical
                    {
                        elementVector.insert(elementVector.begin(), localDirichletElementVector.begin(), localDirichletElementVector.end());
                        elementVector.insert(elementVector.end(), localNeumannElementVector.begin(), localNeumannElementVector.end());
                        nbDirichletElements += localDirichletElementVector.size();
                    }
                }
            }
        }
    }

    return nbDirichletElements;
}


void computeGeometricParam(elementStruct &element,
                           std::map<std::size_t, std::size_t> &nodeIndices,
                           const std::vector<double> &nodeCoords,
                           std::pair<double, double> nodalDisp1,
                           std::pair<double, double> nodalDisp2)
{
    std::size_t i1Tag = element.nodeTags[0];
    std::size_t i2Tag = element.nodeTags[1];

    std::size_t i1Index = nodeIndices[i1Tag];
    std::size_t i2Index = nodeIndices[i2Tag];

    // Get the coordinates of the nodes from their tag.
    element.x1 = nodeCoords[3*i1Index] + nodalDisp1.first;
    element.x2 = nodeCoords[3*i2Index] + nodalDisp2.first;
    element.y1 = nodeCoords[3*i1Index + 1] + nodalDisp1.second;
    element.y2 = nodeCoords[3*i2Index + 1] + nodalDisp2.second;

    element.midX = (element.x1 + element.x2) / 2;
    element.midY = (element.y1 + element.y2) / 2;
    element.deltaX = element.x2 - element.x1;
    element.deltaY = element.y2 - element.y1;
    element.length = sqrt(element.deltaX * element.deltaX + element.deltaY * element.deltaY);
}

void fillSystem(Eigen::MatrixXd &A,
                Eigen::MatrixXd &B,
                Eigen::MatrixXd &c,
                const std::vector<elementStruct> &elementVector,
                const std::size_t nbDirichletElements,
                const std::string integrationType,
                const bool solutionType)
{
    double h;
    double g;
    #pragma omp parallel for collapse(2) schedule(guided) shared(elementVector, A, B, c) private(h, g)
    for(std::size_t i = 0; i < elementVector.size(); i++)
    {
        for(std::size_t j = 0; j < elementVector.size(); j++)
        {
            // Will be computed and stored in A and B.
            h = 0;
            g = 0;

            if(i == j)  // Formula for diagonal elements.
            {
                h = -0.5;
                g = (elementVector[i].length/(2*M_PI)) * (log(elementVector[i].length/2)-1);
            }
            else    // General case
            {
                h = computeAnalyticalH(elementVector[i].midX, elementVector[i].midY, elementVector[j]);
                if(!solutionType)
                    g = computeNumericalG(elementVector[i].midX, elementVector[i].midY, elementVector[j], integrationType);
                else
                    g = computeAnalyticalG(elementVector[i].midX, elementVector[i].midY, elementVector[j]);
            }

            // Storage of h and g at the right place (in matrices A and B).
            if(j < nbDirichletElements)
            {
                A(i,j) = g;
                B(i,j) = h;
                c(j) = elementVector[j].bcValue[0];
            }
            else
            {
                A(i,j) = -h;
                B(i,j) = -g;
                c(j) = elementVector[j].bcValue[1];
            }
        }
    }
}

void computeElectrostaticPressure(std::map<int, double> &electrostaticPressure,
                                  std::vector<elementStruct> &elementVector,
                                  const Eigen::MatrixXd &x,
                                  const std::size_t nbDirichletElements,
                                  const std::string domainName)
{
    // Retrieving the dielectric permittivity of the BEM domain
    std::vector<std::string> keys;
    gmsh::onelab::getNames(keys, "(Materials).+");

    double epsilon = 8.8541878128e-12; // dielectric permittivity
    bool setEpsilon = false;
    for (auto &key : keys)
    {
        // get corresponding value
        std::vector<double> value;
        gmsh::onelab::getNumber(key, value);
        // expected key structure is "type/group_name/field"
        // => split the key string into components
        std::stringstream ss(key);
        std::vector<std::string> words;
        std::string word;
        while(std::getline(ss, word, '/')) // read string until '/'
            words.push_back(word);
        if(words.size()==3)
        {
            if(words[2].compare("Epsilon") == 0 && words[1].compare(domainName) == 0)
            {
                epsilon = value[0];
                setEpsilon = true;
            }
        }
    }
    if(!setEpsilon)
        std::cout << "You forgot to set the dielectric permittivity of the material for the domain '" << domainName << "' in the .geo file and it thus has been fixed to the dielectric permittivity of the vacuum: " << epsilon << " [F/m]\n";

    // Calculates the electrostatic pressure per element in order to provide these results to the FEM code that will handle the rest of the algorithm
    #pragma omp parallel for schedule(guided)
    for(size_t i = 0; i < elementVector.size(); i++)
    {
        if(i < nbDirichletElements)
            elementVector[i].bcValue[1] = x(i);
        else
            elementVector[i].bcValue[0] = x(i);

        double pressure = epsilon * elementVector[i].bcValue[1] * elementVector[i].bcValue[1] / 2; // don't forget factor 2

        #pragma omp critical
        electrostaticPressure[elementVector[i].tag] = pressure;
    }
}


int getModel(std::vector<size_t> &domainNodeTags,
             std::map<std::size_t, std::size_t> &domainNodeIndices,
             std::vector<double> &domainNodeCoords,
             std::vector<int> &domainEntityTags,
             std::map<int, std::size_t> &domainEntityIndices,
             std::vector<std::vector<int>> &domainElementTypes,
             std::vector<std::vector<std::vector<size_t>>> &domainElementTags,
             std::map<std::size_t, std::vector<std::size_t>> &domainElementNodeTags,
             std::map<std::string, std::pair<int, int>> &physicalGroups,
             const std::string domainName)
{
    // Retrieves everything only for the domain of interrest.
    int physicalGroupDim = physicalGroups[domainName].first;
    int physicalGroupTag = physicalGroups[domainName].second;

    gmsh::model::mesh::getNodesForPhysicalGroup(physicalGroupDim, physicalGroupTag, domainNodeTags, domainNodeCoords);

    // The "domainNodeIndices" map is used for finding the position of the coordinates of a node in "domainNodeCoords"
    // from its tag.
    for(std::size_t index = 0; index < domainNodeTags.size(); index++)
        domainNodeIndices[domainNodeTags[index]] = index;

    gmsh::model::getEntitiesForPhysicalGroup(physicalGroupDim, physicalGroupTag, domainEntityTags);
    domainElementTypes.resize(domainEntityTags.size());
    domainElementTags.resize(domainEntityTags.size());
    for(std::size_t j = 0; j < domainEntityTags.size(); j++)
    {
        domainEntityIndices[domainEntityTags[j]] = j;
        
        std::vector<int> elementTypes;
        std::vector<std::vector<std::size_t>> elementTags;
        std::vector<std::vector<std::size_t>> nodeTags;
        gmsh::model::mesh::getElements(elementTypes, elementTags, nodeTags, physicalGroupDim, domainEntityTags[j]);

        domainElementTypes[j].resize(elementTypes.size());
        domainElementTags[j].resize(elementTypes.size());
        for(std::size_t k = 0; k < elementTypes.size(); k++)
        {
            int nbNodes = nodeTags[k].size() / elementTags[k].size();
            std::vector<std::size_t> elementNodeTags(nbNodes);

            domainElementTypes[j][k] = elementTypes[k];
            domainElementTags[j][k].resize(elementTags[k].size());

            for(std::size_t l = 0; l < elementTags[k].size(); l++)
            {
                domainElementTags[j][k][l] = elementTags[k][l];
                int n = 0;
                for(std::size_t m = l*nbNodes; m < (l+1)*nbNodes; m++)
                {
                    elementNodeTags[n] = nodeTags[k][m];
                    n++;
                }
                
                domainElementNodeTags[elementTags[k][l]] = elementNodeTags;
            }
        }
    }

    return physicalGroupTag;
}