#include <gmsh.h>
#include <iostream>
#include <map>
#include <sstream>
#include <Eigen/Dense>
#include <cmath>
#include <iomanip>
#include <stdio.h>
#include <chrono>
#include <omp.h>

// Each element has all its information gathered as in this structure.
struct elementStruct{
    std::size_t tag;            // tag of the element
    double bcValue [2]; // values of [u, u_n]
    size_t nodeTags [2];   // tags of the [first, second] nodes of the element
    size_t midNodeTag;
    double length;
    double x1, y1, x2, y2;
    double midX, midY;
    double deltaX, deltaY;
};

/*
 *  Main
 */

// Fills the "electrostaticPressure" map that gives the electrostatic pressure associated with each (boundary) element tag.
// "nbViews" is the counter of views (useful if other views have to be generated than with this solver). "nodalDisplacements" is
// a map associating the (x,y) displacement to a serie of node tags (if the position of the nodes are not striclty the one
// extracted from the .geo file). "postProcessing" indicates whether the post-processing should be done ('true') or not
// ('false'). "iteration" represents the number of times the solver has been called (some information is printed on the first
// iteration).
void solverBEM(std::map<int, double> &electrostaticPressure,
               int & nbViews,
               std::map<int,std::pair<double,double>> & nodalDisplacements,
               const bool postProcessing,
               const int iteration,
               const bool untangleMesh);

// Fills the "electrostaticPressure" map. The "nodalDisplacements", "postProcessing" and "iteration" arguments are described in
// the description of the function "solverBEM". "domainName" is used for extracting everything in the appropriate independant BEM
// domain, "phiViewTag" and "electricFieldViewTag" are the view tag of the potential and of the electric field respectively.
// "computePhi" and "computeElectricField" indicates whether the potential and/or the electric field should be computed ('true')
// or not ('false').
void singleDomainSolverBEM(std::map<int, double> &electrostaticPressure,
                           const int nbViews,
                           std::map<int,std::pair<double,double>> & nodalDisplacements,
                           const bool postProcessing,
                           const int iteration,
                           const std::string domainName,
                           const int phiViewTag,
                           const int electricFieldViewTag,
                           const bool computePhi,
                           const bool computeElectricField);

/*
 *  Boundary element
 */
void getBC(std::map<std::string, std::pair<bool, double>> &bcMap,
           const std::string domainName);

std::size_t fillElementVector(std::vector<elementStruct> &elementVector,
                      std::map<std::string, std::pair<int, int>> &physicalGroups,
                      std::map<int, std::pair<double, double>> &nodalDisplacements,
                      const std::string domainName);

int getModel(std::vector<size_t> &domainNodeTags,
             std::map<std::size_t, std::size_t> &domainNodeIndices,
             std::vector<double> &domainNodeCoords,
             std::vector<int> &domainEntityTags,
             std::map<int, std::size_t> &domainEntityIndices,
             std::vector<std::vector<int>> &domainElementTypes,
             std::vector<std::vector<std::vector<size_t>>> &domainElementTags,
             std::map<std::size_t, std::vector<std::size_t>> &domainElementNodeTags,
             std::map<std::string, std::pair<int, int>> &physicalGroups,
             const std::string domainName);

void computeGeometricParam(elementStruct &element,
                           std::map<std::size_t, std::size_t> &nodeIndices,
                           const std::vector<double> &nodeCoords,
                           std::pair<double, double> nodalDisp1,
                           std::pair<double, double> nodalDisp2);

void fillSystem(Eigen::MatrixXd &A,
                Eigen::MatrixXd &B,
                Eigen::MatrixXd &c,
                const std::vector<elementStruct> &elementVector,
                const std::size_t nbDirichletElements,
                const std::string integrationType,
                const bool solutionType);

void computeElectrostaticPressure(std::map<int, double> &electrostaticPressure,
                                  std::vector<elementStruct> &elementVector,
                                  const Eigen::MatrixXd &x,
                                  const std::size_t nbDirichletElements,
                                  const std::string domainName);

// Computations
double computeAnalyticalH(const double x, const double y, const elementStruct &element);
double computeNumericalG(const double x, const double y, const elementStruct &element, const std::string integrationType);
double computeNumericalGradHx(const double x, const double y, const elementStruct &element, const std::string integrationType);
double computeNumericalGradHy(const double x, const double y, const elementStruct &element, const std::string integrationType);
double computeNumericalGradGx(const double x, const double y, const elementStruct &element, const std::string integrationType);
double computeNumericalGradGy(const double x, const double y, const elementStruct &element, const std::string integrationType);
double computeAnalyticalGradHx(const double x, const double y, const elementStruct &element);
double computeAnalyticalGradHy(const double x, const double y, const elementStruct &element);
double computeAnalyticalGradGx(const double x, const double y, const elementStruct &element);
double computeAnalyticalGradGy(const double x, const double y, const elementStruct &element);
double computeAnalyticalG(const double x, const double y, const elementStruct &element);

// Post processing
void computeInternalPhi(std::map<int, double> &internalNodalPhi,
                          const std::vector<double> &domainNodeCoords,
                          const std::vector<size_t> &domainNodeTags,
                          std::map<std::size_t, std::size_t> &domainNodeIndices,
                          const std::vector<elementStruct> &elementVector,
                          const std::string integrationType,
                          std::map<std::size_t, std::vector<std::size_t>> &boundaryNodeToElements,
                          const bool &solutionType);

void computeNodeToElements(std::map<std::size_t, std::vector<std::size_t>> &boundaryNodeToElements,
                           const std::vector<elementStruct> &elementVector);

void boundaryElementTags(std::vector<size_t> &domainBoundaryElementTags,
                         const std::vector<std::vector<int>> &domainElementTypes,
                         const std::vector<std::vector<std::vector<size_t>>> &domainElementTags,
                         const std::vector<elementStruct> &elementVector,
                         const int domainPhysicalGroupTag);

void dataInitialization(std::vector<std::vector<std::vector<std::vector<double>>>> &data,
                        std::map<std::size_t, std::vector<std::size_t>> &position,
                        const std::vector<std::vector<int>> &domainElementTypes,
                        const std::vector<std::vector<std::vector<size_t>>> &domainElementTags);

void boundaryElementFinalization(std::vector<std::vector<std::vector<std::vector<double>>>> &data,
                                 const std::vector<std::vector<std::vector<size_t>>> &domainElementTags,
                                 std::map<std::size_t, std::vector<std::size_t>> &domainElementNodeTags,
                                 std::map<int, double> &internalNodalPhi);

void elementNodeData(std::vector<std::vector<std::vector<std::vector<double>>>> &data,
                     const std::vector<std::vector<std::vector<size_t>>> &domainElementTags,
                     const std::vector<std::vector<int>> &domainElementTypes,
                     std::map<std::size_t, std::vector<std::size_t>> &domainElementNodeTags,
                     const std::vector<elementStruct> &elementVector,
                     std::map<int, double> &internalNodalPhi,
                     const int domainPhysicalGroupTag,
                     std::map<std::size_t, std::vector<std::size_t>> &boundaryNodeToElements);

void computeElementData(std::vector<std::vector<std::vector<std::vector<double>>>> &phi,
                        std::vector<std::vector<std::vector<std::vector<double>>>> &electricField,
                        const std::vector<std::vector<std::vector<std::size_t>>> &domainElementTags,
                        const std::vector<std::vector<int>> &domainElementTypes,
                        const std::vector<int> &domainEntityTags,
                        const std::vector<elementStruct> &elementVector,
                        const bool solutionType,
                        const std::string integrationType,
                        const bool computePhi,
                        const bool computeElectricField);

void displayData(const int viewTag,
                 const std::string modelName,
                 const std::string fileName,
                 const std::string dataType,
                 const std::vector<std::vector<std::vector<size_t>>> &domainElementTags,
                 const std::vector<std::vector<std::vector<std::vector<double>>>> &data,
                 const int viewIndex);

// Verification
void dispNodeCoordFun(const std::vector<size_t> &allNodeTags, const std::vector<double> &allCoord);
void displayTime(const std::chrono::time_point<std::chrono::system_clock> start, const std::chrono::time_point<std::chrono::system_clock> end, std::string function_name, bool &showTime);

void convergenceFun(const std::vector<int> &domainEntityTags,
                    const std::vector<std::vector<int>> &domainElementTypes,
                    const std::vector<std::vector<std::vector<size_t>>> &domainElementTags,
                    const std::vector<std::vector<std::vector<std::vector<double>>>> &elementalPhi,
                    const std::vector<std::vector<std::vector<std::vector<double>>>> &elementalElectricField);