#include <gmsh.h>
#include <iostream>
#include <map>
#include <sstream>
#include <Eigen/Dense>
#include <cmath>
#include <iomanip>
#include <stdio.h>
#include <chrono>
#include "functionsBEM.hpp"
#include <omp.h>
#include <algorithm>
#include <list>



/*
 *  Post processing
 */

void computeNodeToElements(std::map<std::size_t, std::vector<std::size_t>> &boundaryNodeToElements,
                           const std::vector<elementStruct> &elementVector)
{
    for(std::size_t i = 0; i < elementVector.size(); i++)
    {
        if(!boundaryNodeToElements.count(elementVector[i].nodeTags[0]))
            boundaryNodeToElements[elementVector[i].nodeTags[0]] = {i};
        else
            boundaryNodeToElements[elementVector[i].nodeTags[0]].push_back(i);

        if(!boundaryNodeToElements.count(elementVector[i].nodeTags[1]))
            boundaryNodeToElements[elementVector[i].nodeTags[1]] = {i};
        else
            boundaryNodeToElements[elementVector[i].nodeTags[1]].push_back(i);

        if(elementVector[i].midNodeTag != 0)
            boundaryNodeToElements[elementVector[i].midNodeTag] = {i};
    }
}

void computeInternalPhi(std::map<int, double> &internalNodalPhi,
                          const std::vector<double> &domainNodeCoords,
                          const std::vector<size_t> &domainNodeTags,
                          std::map<std::size_t, std::size_t> &domainNodeIndices,
                          const std::vector<elementStruct> &elementVector,
                          const std::string integrationType,
                          std::map<std::size_t, std::vector<std::size_t>> &boundaryNodeToElements,
                          const bool &solutionType)
{
    double x = 0;
    double y = 0;
    double phi = 0;
    double h = 0;
    double g = 0;
    std::vector<double> nodeVectorDomain(domainNodeTags.size(), 0);

    #pragma omp parallel for schedule(static) private(x, y, phi, h, g) shared(domainNodeTags, domainNodeCoords, elementVector, internalNodalPhi, boundaryNodeToElements, solutionType)
    for(std::size_t i = 0; i < domainNodeTags.size(); i++)
    {
        if(!boundaryNodeToElements.count(domainNodeTags[i]))
        {
            // Coordinates of the node
            x = domainNodeCoords[3*domainNodeIndices[domainNodeTags[i]]];
            y = domainNodeCoords[3*domainNodeIndices[domainNodeTags[i]] + 1];

            phi = 0;
            for(size_t j = 0; j < elementVector.size(); j++)
            {
                h = computeAnalyticalH(x, y, elementVector[j]);
                if(!solutionType)
                    g = computeNumericalG(x, y, elementVector[j], integrationType);
                else
                    g = computeAnalyticalG(x, y, elementVector[j]);

                phi += h*elementVector[j].bcValue[0] - g*elementVector[j].bcValue[1];
            }
            nodeVectorDomain[i] = phi;
        }
    }

    for(std::size_t i = 0; i < domainNodeTags.size(); i++)
    {
        internalNodalPhi[domainNodeTags[i]] = nodeVectorDomain[i];
    }
}

void elementNodeData(std::vector<std::vector<std::vector<std::vector<double>>>> &data,
                     const std::vector<std::vector<std::vector<size_t>>> &domainElementTags,
                     const std::vector<std::vector<int>> &domainElementTypes,
                     std::map<std::size_t, std::vector<std::size_t>> &domainElementNodeTags,
                     const std::vector<elementStruct> &elementVector,
                     std::map<int, double> &internalNodalPhi,
                     const int domainPhysicalGroupTag,
                     std::map<std::size_t, std::vector<std::size_t>> &boundaryNodeToElements)
{    
    // This map may be replaced in the future. It allows to retain the position of the elements of the elementTag2d loop and linked their position to the loop of the elementTag2d_boundary (the tags of the elements in common with the boundary).
    // This map is the link between the array of tags of the 2d elements of the boundary and that of the tags of all the 2d elements.
    std::map<std::size_t, std::vector<std::size_t>> position;
    dataInitialization(data, position, domainElementTypes, domainElementTags);

    // // The first part of the algorithm consists in recovering the tag of the 2d elements of the domain which have at least one node in common with the boundary
    // // We initialise the first dimension of the vector that will contain all 2d elements in common with the boundary with the number of element types present in the domain.
    std::vector<size_t> domainBoundaryElementTags;
    boundaryElementTags(domainBoundaryElementTags, domainElementTypes, domainElementTags, elementVector, domainPhysicalGroupTag);

    // The second part of the algorithm will consist in assigning the correct potential value to the nodes of the elements in common with the boundary.
    // The first loop is on the elementType in 2d in common with the boundary.
    #pragma omp parallel for schedule(guided)
    for(std::size_t i = 0; i < domainBoundaryElementTags.size(); i++)
    {
        std::size_t elementTag = domainBoundaryElementTags[i];
        int entityTagIndex = position[elementTag][0];
        int elementTypeIndex = position[elementTag][1];
        int elementTagIndex = position[elementTag][2];
        int nbNodes = data[entityTagIndex][elementTypeIndex][elementTagIndex].size();
        std::vector<std::size_t> nodeTags = domainElementNodeTags[elementTag];
        
        // The third loop is on the nodes of a specific elementTag in 2d in common with the boundary.
        for(int j = 0; j < nbNodes; j++)
        {
            std::size_t elementNodeTag = nodeTags[j];

            if(boundaryNodeToElements.count(elementNodeTag))
            {
                std::vector<std::size_t> commonBoundaryElementIndices = boundaryNodeToElements[elementNodeTag];

                // Mid node of a segment on the boundary => take the value of the element.
                if(commonBoundaryElementIndices.size() == 1)
                {
                    data[entityTagIndex][elementTypeIndex][elementTagIndex][j] = elementVector[commonBoundaryElementIndices[0]].bcValue[0];
                }
                else if(commonBoundaryElementIndices.size() == 2)
                {
                    double phi = 0.0;
                    int count = 0;

                    // First element.
                    int adjacentElementIndex = commonBoundaryElementIndices[0];
                    std::size_t adjacentNodeTag = elementVector[adjacentElementIndex].nodeTags[0];
                    if(adjacentNodeTag == elementNodeTag)
                    {
                        adjacentNodeTag = elementVector[adjacentElementIndex].nodeTags[1];
                    }

                    // if 'adjacentNodeTag' is in 'nodeTags'
                    if(std::count(nodeTags.begin(), nodeTags.end(), adjacentNodeTag))
                    {
                        count++;
                    }
                    phi += elementVector[adjacentElementIndex].bcValue[0];


                    // Second element.
                    adjacentElementIndex = commonBoundaryElementIndices[1];
                    adjacentNodeTag = elementVector[adjacentElementIndex].nodeTags[0];
                    if(adjacentNodeTag == elementNodeTag)
                    {
                        adjacentNodeTag = elementVector[adjacentElementIndex].nodeTags[1];
                    }

                    // if 'adjacentNodeTag' is in 'nodeTags'
                    if(std::count(nodeTags.begin(), nodeTags.end(), adjacentNodeTag))
                    {
                        if(count == 0)
                        {
                            phi = elementVector[adjacentElementIndex].bcValue[0];
                        }
                        else
                        {
                            phi += elementVector[adjacentElementIndex].bcValue[0];
                            phi /= 2;
                        }
                    }
                    else
                    {
                        if(count == 0)
                        {
                            phi += elementVector[adjacentElementIndex].bcValue[0];
                            phi /= 2;
                        }
                    }

                    data[entityTagIndex][elementTypeIndex][elementTagIndex][j] = phi;
                }
            }
        }
    }

    //The last step is to loop over all the 2d elements in the domain. If the HUGE_VAL value is still present, this means that the element node does not belong to the boundary (since the value of the potential has not been modified) and is therefore part of the domain. We then use the map which associates the value of the potential with the tag of the node inside the domain.
    boundaryElementFinalization(data, domainElementTags, domainElementNodeTags, internalNodalPhi);
}


void dataInitialization(std::vector<std::vector<std::vector<std::vector<double>>>> &data,
                        std::map<std::size_t, std::vector<std::size_t>> &position,
                        const std::vector<std::vector<int>> &domainElementTypes,
                        const std::vector<std::vector<std::vector<size_t>>> &domainElementTags)
{
    data.resize(domainElementTypes.size());
    for(std::size_t i = 0; i < domainElementTypes.size(); i++)
    {
        data[i].resize(domainElementTypes[i].size());
        
        #pragma omp parallel for schedule(guided) shared(data, domainElementTypes, domainElementTags, position)
        for(std::size_t j = 0; j < domainElementTypes[i].size(); j++)
        {
            std::string elementName;
            int dim;
            int order;
            int nbNodes;
            std::vector<double> localNodeCoord;
            int nbPrimaryNodes;
            gmsh::model::mesh::getElementProperties(domainElementTypes[i][j], elementName, dim, order, nbNodes, localNodeCoord, nbPrimaryNodes);

            data[i][j].resize(domainElementTags[i][j].size());
            for(std::size_t k = 0; k < domainElementTags[i][j].size(); k++)
            {
                data[i][j][k].resize(nbNodes, HUGE_VAL);
                std::vector<std::size_t> pos = {i, j, k};

                #pragma omp critical
                    position[domainElementTags[i][j][k]] = pos;
            }
        }
    }
}

void boundaryElementTags(std::vector<std::size_t> &domainBoundaryElementTags,
                         const std::vector<std::vector<int>> &domainElementTypes,
                         const std::vector<std::vector<std::vector<size_t>>> &domainElementTags,
                         const std::vector<elementStruct> &elementVector,
                         const int domainPhysicalGroupTag)
{
    #pragma omp parallel
    {
        std::vector<std::size_t> localElementTags;

        // This loop goes through all the nodes of the boundary (node 1 of each element). We then use getElementsByCoordinates to obtain all the elements in common with the node considered. We only take the 2d elements.
        #pragma omp for schedule(guided)
        for(std::size_t i = 0; i < elementVector.size(); i++)
        {
            std::vector<size_t> elementTags;
            gmsh::model::mesh::getElementsByCoordinates(elementVector[i].x1, elementVector[i].y1, 0, elementTags, 2);

            // We loop on the 2d elements which are in common with the considered node of the boundary. Depending on their type, they are added to the vector comprising all the 2d elements in contact with the boundary.
            for(std::size_t j = 0; j < elementTags.size(); j++)
            {
                int elementType;
                std::vector<size_t> nodeTags;
                int entityDim;
                int entityTag;
                std::vector<int> physicalTags;
                #pragma omp critical
                {
                    gmsh::model::mesh::getElement(elementTags[j], elementType, nodeTags, entityDim, entityTag);
                    gmsh::model::getPhysicalGroupsForEntity(entityDim, entityTag, physicalTags);
                }
                
                for(std::size_t k = 0; k < physicalTags.size(); k++)
                    if(physicalTags[k] == domainPhysicalGroupTag)
                        localElementTags.push_back(elementTags[j]);
            }
        }

        #pragma omp critical
        domainBoundaryElementTags.insert(domainBoundaryElementTags.end(), localElementTags.begin(), localElementTags.end());
    }
    

    // Allows duplicate tags to be removed. If an element has several nodes with the border, it has been counted several times before.
    sort( domainBoundaryElementTags.begin(), domainBoundaryElementTags.end() );
    domainBoundaryElementTags.erase( unique( domainBoundaryElementTags.begin(), domainBoundaryElementTags.end() ), domainBoundaryElementTags.end() );
}


void boundaryElementFinalization(std::vector<std::vector<std::vector<std::vector<double>>>> &data,
                                 const std::vector<std::vector<std::vector<size_t>>> &domainElementTags,
                                 std::map<std::size_t, std::vector<std::size_t>> &domainElementNodeTags,
                                 std::map<int, double> &internalNodalPhi)
{
    for(std::size_t i = 0; i < domainElementTags.size(); i++)
    {
        for(std::size_t j = 0; j < domainElementTags[i].size(); j++)
        {
            #pragma omp parallel for schedule(guided)
            for(std::size_t k = 0; k < domainElementTags[i][j].size(); k++)
            {
                std::vector<std::size_t> nodeTags = domainElementNodeTags[domainElementTags[i][j][k]];
                for(std::size_t l = 0; l < nodeTags.size(); l++)
                {
                    std::size_t elementNodeTag = nodeTags[l];
                    if(data[i][j][k][l] == HUGE_VAL && internalNodalPhi.count(elementNodeTag))
                        data[i][j][k][l] = internalNodalPhi[elementNodeTag];
                }
            }
        }
    }
}


void computeElementData(std::vector<std::vector<std::vector<std::vector<double>>>> &phi,
                        std::vector<std::vector<std::vector<std::vector<double>>>> &electricField,
                        const std::vector<std::vector<std::vector<std::size_t>>> &domainElementTags,
                        const std::vector<std::vector<int>> &domainElementTypes,
                        const std::vector<int> &domainEntityTags,
                        const std::vector<elementStruct> &elementVector,
                        const bool solutionType,
                        const std::string integrationType,
                        const bool computePhi,
                        const bool computeElectricField)
{
    std::vector<double> barycenters;

    phi.resize(domainEntityTags.size());
    electricField.resize(domainEntityTags.size());
    for(std::size_t i = 0; i < domainEntityTags.size(); i++)
    {   
        phi[i].resize(domainElementTypes[i].size());
        electricField[i].resize(domainElementTypes[i].size());
        for(std::size_t j = 0; j < domainElementTypes[i].size(); j++)
        {
            phi[i][j].resize(domainElementTags[i][j].size());
            electricField[i][j].resize(domainElementTags[i][j].size());
            gmsh::model::mesh::getBarycenters(domainElementTypes[i][j], domainEntityTags[i], false, false, barycenters);
            for(size_t k = 0; k < domainElementTags[i][j].size(); k++)
            {
                double x = barycenters[3*k];
                double y = barycenters[3*k + 1];

                double currentPhi = 0;
                double gradPhiX = 0;
                double gradPhiY = 0;

                #pragma omp parallel shared(currentPhi, gradPhiX, gradPhiY)
                {
                    double localCurrentPhi = 0;
                    double localGradPhiX = 0;
                    double localGradPhiY = 0;

                    double h, hX, hY;
                    double g, gX, gY;
                    
                    #pragma omp for schedule(guided) 
                    for(std::size_t el = 0; el < elementVector.size(); el++)
                    {
                        if(computePhi)
                        {
                            h = computeAnalyticalH(x, y, elementVector[el]);
                            if(!solutionType)
                                g = computeNumericalG(x, y, elementVector[el], integrationType);
                            else
                                g = computeAnalyticalG(x, y, elementVector[el]);

                            localCurrentPhi += h*elementVector[el].bcValue[0] - g*elementVector[el].bcValue[1];
                        }
                        if(computeElectricField)
                        {
                            if(!solutionType)
                            {
                                hX = computeNumericalGradHx(x, y, elementVector[el], integrationType);
                                gX = computeNumericalGradGx(x, y, elementVector[el], integrationType);
                                hY = computeNumericalGradHy(x, y, elementVector[el], integrationType);
                                gY = computeNumericalGradGy(x, y, elementVector[el], integrationType);
                            }
                            else
                            {
                                hX = computeAnalyticalGradHx(x, y, elementVector[el]);
                                gX = computeAnalyticalGradGx(x, y, elementVector[el]);
                                hY = computeAnalyticalGradHy(x, y, elementVector[el]);
                                gY = computeAnalyticalGradGy(x, y, elementVector[el]);
                            }

                            localGradPhiX += hX*elementVector[el].bcValue[0] - gX*elementVector[el].bcValue[1];
                            localGradPhiY += hY*elementVector[el].bcValue[0] - gY*elementVector[el].bcValue[1];
                        }
                    }

                    #pragma omp critical
                    {
                        currentPhi += localCurrentPhi;
                        gradPhiX += localGradPhiX;
                        gradPhiY += localGradPhiY;
                    }
                }

                phi[i][j][k].resize(1, currentPhi);
                electricField[i][j][k].resize(3);
                electricField[i][j][k][0] = -gradPhiX;
                electricField[i][j][k][1] = -gradPhiY;
                electricField[i][j][k][2] = 0.0;
            }
        }
    }
}

void displayData(const int viewTag,
                 const std::string modelName,
                 const std::string fileName,
                 const std::string dataType,
                 const std::vector<std::vector<std::vector<std::size_t>>> &domainElementTags,
                 const std::vector<std::vector<std::vector<std::vector<double>>>> &data,
                 const int viewIndex)
{
    for(std::size_t i = 0; i < data.size(); i++)
        for(std::size_t j = 0; j < data[i].size(); j++)
            gmsh::view::addModelData(viewTag, 0 , modelName, dataType, domainElementTags[i][j], data[i][j]);
    
    gmsh::view::write(viewTag, fileName);
    
    gmsh::option::setNumber("View[" + std::to_string(viewIndex) + "].IntervalsType", 3);
    gmsh::option::setNumber("View[" + std::to_string(viewIndex) + "].MaxRecursionLevel", 3);
    gmsh::option::setNumber("View[" + std::to_string(viewIndex) + "].TargetError", -0.0001);
}